Imports System
Imports System.Windows.Forms
Imports System.Runtime.InteropServices
Imports System.Xml
Imports CQG

Module modAlgorithmicOrder

   ' Declare the SetConsoleCtrlHandler function as external and receiving a delegate.   
   <DllImport("Kernel32")> _
   Private Function SetConsoleCtrlHandler(ByVal Handler As HandlerRoutine, _
        ByVal Add As Boolean) As Boolean
   End Function

   ' A delegate type to be used as the handler routine for SetConsoleCtrlHandler.
   Delegate Function HandlerRoutine(ByVal CtrlType As eCtrlTypes) As Boolean

   ' Flag indicates whether algorithmic order is registered.
   Private m_IsAlgorithmicOrderRegistered As Boolean

   ' CQGCEL object.
   Private WithEvents m_CEL As CQGCEL

   ' Placed orders count.
   Private m_OrdersCount As Integer

   ' Main account id.
   Private m_MainAccountID As Integer

   Private Structure AccountProperties

      Public AccountID As Integer
      Public ReplicationRatio As Integer

   End Structure

   ' Collection of replication accounts ids and ratios.
   Private m_RepAccountProps As New ArrayList

   ' Flag indicating whether to multiply order quantity by ratio
   Private m_multiplyByRatio As Boolean

   Sub Main()
      Console.WriteLine("This application will register the REP order type in CQG IC.")
      Console.WriteLine()
      Console.WriteLine("After the registration, you will be able to place REP orders from within")
      Console.WriteLine("the CQG IC. Each REP order placed under main account will be replicated ")
      Console.WriteLine("for each selected account. Please note that all activity is done through ")
      Console.WriteLine("this application. Hence the replication will not take place if the appli-")
      Console.WriteLine("cation is not running. All cancellations and modifications should be done ")
      Console.WriteLine("manually for each order on each account. ")
      Console.WriteLine()
      Console.WriteLine("The REP order type will be automatically unregistered when this console")
      Console.WriteLine("application is terminated.")
      Console.WriteLine("_________________________________________________________")

      Try
         ' Initialize private variables
         m_OrdersCount = 0
         m_IsAlgorithmicOrderRegistered = False

         InitializeCQGCEL()
         InitializeData()

         RunApplication()

         If m_IsAlgorithmicOrderRegistered Then
            UnRegisterAlgorithmicOrder()
         End If

         m_CEL.Shutdown()
      Catch ex As Exception
         Console.WriteLine("Error: " & ex.Message)
      End Try

   End Sub

   Private Sub RunApplication()

      ' Use interop to set a console control handler
      Dim hr As New modAlgorithmicOrder.HandlerRoutine(AddressOf Handler)
      modAlgorithmicOrder.SetConsoleCtrlHandler(hr, True)

      ' The object hr is not referred to again.
      ' The garbage collector can detect that the object has no
      ' more managed references and might clean it up here while
      ' the unmanaged SetConsoleCtrlHandler method is still using it.     
      ' Force a garbage collection to demonstrate how the hr
      ' object will be handled.
      GC.Collect()
      GC.WaitForPendingFinalizers()
      GC.Collect()

      Application.Run()

      ' Call GC.KeepAlive(hr) at this point to maintain a reference to hr. 
      ' This will prevent the garbage collector from collecting the 
      ' object during the execution of the SetConsoleCtrlHandler method.
      GC.KeepAlive(hr)

   End Sub

   ''' <summary>
   ''' Initialize CQGCEL object and start up it.
   ''' </summary>
   Private Sub InitializeCQGCEL()

      Console.Write("Initializing CQGCEL...")

      ' Creating new CQGCEL object
      m_CEL = New CQGCEL

      m_CEL.APIConfiguration.ReadyStatusCheck = CQG.eReadyStatusCheck.rscOff
      m_CEL.APIConfiguration.CollectionsThrowException = False
      m_CEL.APIConfiguration.UseOrderSide = True
      m_CEL.APIConfiguration.TimeZoneCode = CQG.eTimeZone.tzCentral

      ' Starting up the CQGCEL
      m_CEL.Startup()

   End Sub

   ''' <summary>
   ''' Loads accounts ids and working mode from the xml file.
   ''' </summary>
   Private Sub InitializeData()

      Try

         Dim accDoc As XmlDocument = New Xml.XmlDocument
         accDoc.Load(Application.StartupPath & "\\" & XML_FILE_NAME)

         Dim mainAccNode As XmlNode = accDoc.SelectSingleNode("/accounts/main_account")
         m_MainAccountID = CInt(mainAccNode.InnerText)

         Dim multiplyByRatioNode As XmlNode = accDoc.SelectSingleNode("/accounts/multiply_by_ratio")
         m_multiplyByRatio = Boolean.Parse(multiplyByRatioNode.InnerText)

         Dim accNodes As XmlNodeList = accDoc.SelectNodes("/accounts/rep_accounts/rep_account")

         Dim index As Integer = 0
         For Each accNode As XmlNode In accNodes

            If Not accNode.HasChildNodes Then
               Return
            End If

            Dim accIdNode As XmlNode = accNode.ChildNodes(0)
            Dim accRatioNode As XmlNode = accNode.ChildNodes(1)

            Dim accProp As AccountProperties = New AccountProperties
            accProp.AccountID = CInt(accIdNode.InnerText)
            accProp.ReplicationRatio = CInt(accRatioNode.InnerText)
            m_RepAccountProps.Add(accProp)
            index += 1
         Next

      Catch ex As Exception
         Console.WriteLine("Error: " & ex.Message)
      End Try

   End Sub

   ''' <summary>
   ''' CQGCEL's CELStarted event handler. Received if CQGCEL started success.
   ''' </summary>
   Private Sub CEL_CELStarted() Handles m_CEL.CELStarted

      Console.WriteLine("OK")
      Console.Write("Connecting to GW...")

   End Sub

   ''' <summary>
   ''' This event is fired, when CQGCEL detects some abnormal discrepancy between data expected and data received.
   ''' </summary>
   ''' <param name="cqg_error">
   ''' Object, in which the error occurred.
   ''' </param>
   ''' <param name="error_description">
   ''' String, describing the error.
   ''' </param>
   Private Sub CEL_DataError( _
               ByVal cqg_error As Object, _
               ByVal error_description As String) Handles m_CEL.DataError

      Console.WriteLine()
      Console.WriteLine("Data Error - " & error_description)

   End Sub

   ''' <summary>
   ''' CQGCEL's AlgorithmicOrderPlaced event handler. Received if algorithmic order placed.
   ''' </summary>
   ''' <param name="guid_">
   ''' Order's guid
   ''' </param>
   ''' <param name="main_params">
   ''' Order's main parameters
   ''' </param>
   ''' <param name="algorithmic_props">
   ''' Order's properties
   ''' </param>
   Private Sub CEL_AlgorithmicOrderPlaced( _
               ByVal guid_ As String, _
               ByVal main_params As CQG.CQGAlgorithmicOrderParameters, _
               ByVal algorithmic_props As CQG.CQGAlgorithmicOrderProperties) Handles m_CEL.AlgorithmicOrderPlaced

      Try
         m_OrdersCount += 1
         If m_OrdersCount > MAX_ORDERS_COUNT Then
            m_OrdersCount = 0
         End If
         ' Check GW connection status
         If m_CEL.Environment.GWConnectionStatus <> CQG.eConnectionStatus.csConnectionUp Then
            Exit Sub
         End If

         ' If main account not selected do nothing
         If m_MainAccountID <> main_params.GWAccountID Then
            Console.WriteLine("Info: Main account GW ID is {0}. Order was placed for {1}", m_MainAccountID, main_params.GWAccountID)
            Exit Sub
         End If

         Dim account As CQG.CQGAccount = m_CEL.Accounts.Item(m_MainAccountID)

         ' Check account
         If account Is Nothing Then
            Console.WriteLine("Error: Invalid account GW ID")
            Exit Sub
         End If

         Dim orderNumber As Integer = 0

         ' Get prices
         Dim limitPrice As Double = 0.0
         Dim stopPrice As Double = 0.0
         Select Case main_params.Type
            Case CQG.eOrderType.otLimit
               limitPrice = main_params.Price
            Case CQG.eOrderType.otStop
               stopPrice = main_params.Price
            Case CQG.eOrderType.otStopLimit
               limitPrice = main_params.StopLimitPrice
               stopPrice = main_params.Price
         End Select

         Dim ueName As String = REP_UENAME & m_OrdersCount.ToString("000000") & orderNumber.ToString("000")

         ' Create order
         Dim order As CQG.CQGOrder = CreateOrder(account, limitPrice, stopPrice, main_params, ueName)

         ' Place order in CQGIC
         order.Place()

         ' Place replication orders
         Dim mainOrderGuid As String = order.GUID

         If m_multiplyByRatio Then
            PlaceRepOrdersMultipliedByRatio(mainOrderGuid, limitPrice, stopPrice, main_params)
         Else
            PlaceRepOrders(mainOrderGuid, limitPrice, stopPrice, main_params)
         End If

      Catch ex As Exception
         Console.WriteLine("Error: " & ex.Message)
      End Try

   End Sub

   ''' <summary>
   ''' Places orders on replication accounts multiplying by order ratio
   ''' </summary>
   ''' <param name="mainOrderGUID">
   ''' Main order's guid
   ''' </param>
   ''' <param name="limitPrice">
   ''' Order's limit price
   ''' </param>
   ''' <param name="stopPrice">
   ''' Order's stop price
   ''' </param>
   ''' <param name="main_params">
   ''' Order's main parameters
   ''' </param>
   Private Sub PlaceRepOrdersMultipliedByRatio( _
               ByVal mainOrderGUID As String, _
               ByVal limitPrice As Double, _
               ByVal stopPrice As Double, _
               ByVal main_params As CQG.CQGAlgorithmicOrderParameters)

      For Each accProp As AccountProperties In m_RepAccountProps

         Dim account As CQGAccount = m_CEL.Accounts.Item(accProp.AccountID)

         ' Check account
         If account Is Nothing Then
            Console.WriteLine("Error: Invalid account GW ID - {0}.", accProp.AccountID)
         Else

            Dim repRatio As Integer = accProp.ReplicationRatio

            ' Make UEName for partial order
            Dim ueName As String = REP_UENAME & m_OrdersCount.ToString("000000") & repRatio.ToString("000")

            ' Create order
            Dim order As CQG.CQGOrder = CreateOrder(account, limitPrice, stopPrice, main_params, ueName, mainOrderGUID, repRatio)

            ' Place order in CQGIC
            order.Place()

         End If
      Next

   End Sub

   ''' <summary>
   ''' Places orders on replication accounts
   ''' </summary>
   ''' <param name="mainOrderGUID">
   ''' Main order's guid
   ''' </param>
   ''' <param name="limitPrice">
   ''' Order's limit price
   ''' </param>
   ''' <param name="stopPrice">
   ''' Order's stop price
   ''' </param>
   ''' <param name="main_params">
   ''' Order's main parameters
   ''' </param>
   Private Sub PlaceRepOrders( _
               ByVal mainOrderGUID As String, _
               ByVal limitPrice As Double, _
               ByVal stopPrice As Double, _
               ByVal main_params As CQG.CQGAlgorithmicOrderParameters)

      For Each accProp As AccountProperties In m_RepAccountProps

         Dim account As CQGAccount = m_CEL.Accounts.Item(accProp.AccountID)

         ' Check account
         If account Is Nothing Then
            Console.WriteLine("Error: Invalid account GW ID - {0}.", accProp.AccountID)
         Else

            Dim orderNumber As Integer = 0
            Dim order As CQGOrder
            Dim ueName As String

            For curOrderNumber As Integer = 1 To accProp.ReplicationRatio

               orderNumber += 1

               ' Make UEName for partial order
               ueName = REP_UENAME & m_OrdersCount.ToString("000000") & orderNumber.ToString("000")

               order = CreateOrder(account, limitPrice, stopPrice, main_params, ueName, mainOrderGUID)

               ' Place order in CQGIC
               order.Place()

            Next curOrderNumber

         End If

      Next

   End Sub

   ''' <summary> 
   ''' Creates an order.
   ''' </summary>
   ''' <param name="account">
   ''' The account.
   ''' </param>
   ''' <param name="limitPrice">
   ''' Order's limit price.
   ''' </param>
   ''' <param name="stopPrice">
   ''' Order's stop price.
   ''' </param>
   ''' <param name="main_params">
   ''' Order's main parameters.
   ''' </param>
   ''' <param name="ueName">
   ''' Name of the ue.
   ''' </param>
   ''' <param name="mainOrderGUID">
   ''' Main order's guid.
   ''' </param>
   ''' <param name="repRatio">
   ''' (optional) the rep ratio.
   ''' </param>
   ''' <returns>
   ''' The new order.
   ''' </returns>
   Private Function CreateOrder( _
                                ByVal account As CQGAccount, _
                                ByVal limitPrice As Double, _
                                ByVal stopPrice As Double, _
                                ByVal main_params As CQG.CQGAlgorithmicOrderParameters, _
                                ByVal ueName As String, _
                                Optional ByVal mainOrderGUID As String = "", _
                                Optional ByVal repRatio As Integer = 1) As CQG.CQGOrder

      Dim isSyntheticInstrument As Boolean = (main_params.ExecutionPattern <> String.Empty)

      ' Create order
      Dim order As CQG.CQGOrder
      If isSyntheticInstrument Then
         order = m_CEL.CreateStrategyOrderByExpression( _
                                    main_params.Type, _
                                    main_params.InstrumentName, _
                                    account, _
                                    Nothing, _
                                    main_params.Quantity * repRatio, _
                                    main_params.Side, _
                                    limitPrice, _
                                    stopPrice, _
                                    ueName)
      Else
         order = m_CEL.CreateOrderByInstrumentName( _
                                       main_params.Type, _
                                       main_params.InstrumentName, _
                                       account, _
                                       main_params.Quantity * repRatio, _
                                       main_params.Side, _
                                       limitPrice, _
                                       stopPrice, _
                                       ueName)
      End If
      If mainOrderGUID <> String.Empty Then
         order.Tag("MAOGUID") = mainOrderGUID
      End If

      'Set duration
      SetDuration(order, main_params)

      ' Set execution instructions
      SetExecutionInstructions(order, main_params)

      'Set speculation type and open close instruction.
      order.Properties(eOrderProperty.opSpeculationType).Value = main_params.SpeculationType
      order.Properties(eOrderProperty.opOpenCloseInstruction).Value = main_params.OpenCloseInstruction

      Dim algoSupplier As CQGAlgoSupplier.CQGAlgoSupplier = TryCast(m_CEL, CQGAlgoSupplier.CQGAlgoSupplier)
      If algoSupplier IsNot Nothing Then
            algoSupplier.SetOrderMiFIDAlgorithmID(order, "CQGAPIReplicator001")
      End If

      Return order

   End Function

   ''' <summary> 
   ''' Registers REP algorithmic type order in CQGIC.
   ''' </summary>
   Private Sub RegisterAlgorithmicOrder()

      Dim aoDefinitions As CQG.CQGAlgorithmicOrderDefinitions = m_CEL.CreateAlgorithmicOrderDefinitions()

      ' Registering REP order
      m_CEL.RegisterAlgorithmicOrder( _
                  REP_GUID, _
                  "Replication", _
                  "RREP", _
                  "Replicates order from one account to multiple accounts.", _
                  aoDefinitions, _
                  1)

   End Sub

   ''' <summary>
   ''' Unregisters REP algorithmic order from CQGIC.
   ''' </summary>
   Private Sub UnRegisterAlgorithmicOrder()

      ' UnRegistering REP order
      Console.WriteLine("Unregistering REP algorithmic order...")
      m_CEL.UnregisterAlgorithmicOrder(REP_GUID)

   End Sub

   ''' <summary>
   ''' This event is fired, when algorithmic order registration is completed.
   ''' </summary>
   ''' <param name="guid_">
   ''' Order's guid
   ''' </param>
   ''' <param name="cqg_error">
   ''' Error object
   ''' </param>
   Private Sub CEL_AlgorithmicOrderRegistrationComplete( _
               ByVal guid_ As String, _
               ByVal cqg_error As CQG.CQGError) Handles m_CEL.AlgorithmicOrderRegistrationComplete

      Try
         If Not m_CEL.IsValid(cqg_error) Then
            Console.WriteLine("OK")
            m_IsAlgorithmicOrderRegistered = True
            Console.Write("Getting Accounts...")
            m_CEL.AccountSubscriptionLevel = CQG.eAccountSubscriptionLevel.aslAccountUpdatesAndOrders
         Else
            Console.WriteLine("Error : " & cqg_error.Description)
         End If
      Catch ex As Exception
         Console.WriteLine("Error: " & ex.Message)
      End Try

   End Sub

   ''' <summary>
   ''' This event is fired when the status of the connection with the CQG Gateway is changed.
   ''' </summary>
   ''' <param name="new_status">
   ''' The current status of the connection with CQG Gateway.
   ''' </param>
   Private Sub CEL_GWConnectionStatusChanged( _
               ByVal new_status As CQG.eConnectionStatus) Handles m_CEL.GWConnectionStatusChanged

      Try
         If new_status = CQG.eConnectionStatus.csConnectionUp Then
            Console.WriteLine("OK")
            If (Not m_IsAlgorithmicOrderRegistered) Then
               Console.Write("Registering REP algorithmic order...")
               RegisterAlgorithmicOrder()
            Else
               Console.Write("Getting Accounts...")
               m_CEL.AccountSubscriptionLevel = CQG.eAccountSubscriptionLevel.aslAccountUpdatesAndOrders
            End If
         ElseIf new_status = CQG.eConnectionStatus.csConnectionDown Then
            Console.WriteLine("GW Connection is Down, please log on to GW from CQGIC")
            Console.Write("Waiting for GW connection...")
         End If
      Catch ex As Exception
         Console.WriteLine("Error: " & ex.Message)
      End Try
   End Sub

   ''' <summary>
   ''' This event is fired, when the account or position information is changed.
   ''' </summary>
   ''' <param name="change_type">
   ''' Account change type, which allows to differentiate the occurred changes. 
   ''' </param>
   ''' <param name="cqg_account">
   ''' A CQGAccount object, representing the account to which the current change refers.
   ''' </param>
   ''' <param name="cqg_position">
   ''' A CQGPosition object, representing the position to which the current change refers.
   ''' </param>
   Private Sub CEL_AccountChanged( _
               ByVal change_type As CQG.eAccountChangeType, _
               ByVal cqg_account As CQG.CQGAccount, _
               ByVal cqg_position As CQG.CQGPosition) Handles m_CEL.AccountChanged

      Try
         If change_type = CQG.eAccountChangeType.actAccountsReloaded Then
            Console.WriteLine("OK")
            Console.WriteLine("Now you can place algorithmic orders...")
         End If
      Catch ex As Exception
         Console.WriteLine("Error: " & ex.Message)
      End Try

   End Sub

   ''' <summary>
   ''' Configures execution instructions and corresponding properties for each instruction.
   ''' </summary>
   ''' <param name="order">
   ''' Order which should be configured. 
   ''' </param>
   ''' <param name="main_params">
   ''' AOB main parameters, which are used to retrieve instructions.
   ''' </param>
   Private Sub SetExecutionInstructions(ByVal order As CQG.CQGOrder, ByVal main_params As CQG.CQGAlgorithmicOrderParameters)
      If main_params.ExecutionPattern <> String.Empty Then
         order.Properties(CQG.eOrderProperty.opExecutionPattern).Value = main_params.ExecutionPattern
      End If

      If main_params.ExecutionInstructions = CQG.eOrderExecutionInstruction.oeiUndefined Then
         Exit Sub
      End If

      order.Properties(CQG.eOrderProperty.opExecutionInstructions).Value = main_params.ExecutionInstructions

      If (main_params.ExecutionInstructions And _
           CQG.eOrderExecutionInstruction.oeiIceberg) = eOrderExecutionInstruction.oeiIceberg Then
         order.Properties(CQG.eOrderProperty.opVisibleQuantity).Value = main_params.VisibleQuantity
      End If

      If (main_params.ExecutionInstructions And _
           CQG.eOrderExecutionInstruction.oeiQTS) = eOrderExecutionInstruction.oeiQTS Then
         order.Properties(CQG.eOrderProperty.opTriggerQuantity).Value = main_params.TriggerQuantity
      End If

      If (main_params.ExecutionInstructions And _
           CQG.eOrderExecutionInstruction.oeiTrailing) = eOrderExecutionInstruction.oeiTrailing Then
         order.Properties(CQG.eOrderProperty.opTrailingOffset).Value = main_params.TrailingOffset
         If order.Type = CQG.eOrderType.otLimit Then
            order.Properties(CQG.eOrderProperty.opPreferredTrailingPeg).Value = main_params.PreferredTrailingPeg
         End If
      End If
   End Sub

   ''' <summary>
   ''' Configures duration of order.
   ''' </summary>
   ''' <param name="order">
   ''' Order which should be configured. 
   ''' </param>
   ''' <param name="main_params">
   ''' AOB main parameters, which are used to retrieve duration.
   ''' </param>
   Private Sub SetDuration(ByVal order As CQG.CQGOrder, ByVal main_params As CQG.CQGAlgorithmicOrderParameters)
      order.DurationType = main_params.DurationType
      If main_params.DurationType = CQG.eOrderDuration.odGoodTillDate OrElse main_params.DurationType = CQG.eOrderDuration.odGoodTillTime Then
         order.Properties(CQG.eOrderProperty.opGTDTime).Value = main_params.GTD
      End If
   End Sub

   Private Function Handler( _
               ByVal CtrlType As eCtrlTypes) As Boolean

      ' Handling the event type

      Select Case CtrlType
         Case eCtrlTypes.ctCtrlCEvent, _
             eCtrlTypes.ctCtrlBreakEvent, _
             eCtrlTypes.ctCloseEvent
            Application.Exit()
      End Select

      Return True

   End Function

   ''' <summary>
   ''' This event is fired when a new order is added to the collection, the order status is changed,
   ''' or the order is removed from the collection.
   ''' </summary>
   ''' <param name="change_type">
   ''' The type of the occurred change.
   ''' </param>
   ''' <param name="cqg_order ">
   ''' CQGOrder object representing the order to which the change refers.
   ''' </param>
   ''' <param name="old_properties">
   ''' CQGOrderProperties collection representing old values of all order properties.
   ''' </param>
   ''' <param name="cqg_fill">
   ''' CQGFill object representing the last fill of the order, i.e. the last item of order's Fills collection.
   ''' </param>
   ''' <param name="cqg_error">
   ''' CQGError object representing either an error sent by CQG Gateway or an intermediate error.
   ''' </param>
   Private Sub CEL_OrderChanged( _
                ByVal change_type As CQG.eChangeType, _
                ByVal cqg_order As CQG.CQGOrder, _
                ByVal old_properties As CQG.CQGOrderProperties, _
                ByVal cqg_fill As CQG.CQGFill, _
                ByVal cqg_error As CQG.CQGError) Handles m_CEL.OrderChanged

      Try

         If cqg_order.Account.GWAccountID = m_MainAccountID AndAlso cqg_order.GWStatus = eOrderStatus.osCanceled Then

            Dim allRepAccountIDs As New Hashtable(m_RepAccountProps.Count - 1)

            Dim j As Integer = 0
            For Each accProp As AccountProperties In m_RepAccountProps
               allRepAccountIDs.Add(j, accProp.AccountID)
               j += 1
            Next

            For Each curOrder As CQGOrder In m_CEL.Orders

               ' In case of cancellation failure pass to the next order
               Try

                  If curOrder.CanBeCanceled AndAlso _
                    allRepAccountIDs.ContainsValue(curOrder.Account.GWAccountID) AndAlso _
                       CStr(curOrder.Tag("MAOGUID")) = cqg_order.GUID Then
                     curOrder.Cancel()
                  End If

               Catch ex As Exception
                  Console.WriteLine("Error: " & ex.Message)
               End Try

            Next

         End If

      Catch ex As Exception
         Console.WriteLine("Error: " & ex.Message)
      End Try

   End Sub

End Module